from .baseline import FirstN
from .graph_text_rank import TextRank, SingleRank, PositionRank, TFIDFRank, LexRank, SingleTPR
from .graph_tpr import TopicalPageRank
from .graph_topic_rank import TopicRank
from .tfidf import TFIDF
from .lexical_specificity import LexSpec, TF
from .lda import LDA
from ._get_dataset import get_benchmark_dataset, VALID_DATASET_LIST
from ._phrase_constructor import PhraseConstructor
from ._get_algorithm import get_algorithm, VALID_ALGORITHMS
from ._stopwords import get_stopwords_list



# """ Implementation of MultipartiteRank """
# import networkx as nx
# import numpy as np
# from itertools import chain
# from .graph_topic_rank import TopicRank
#
# __all__ = "MultipartiteRank"
#
#
# class MultipartiteRank(TopicRank):
#     """ MultipartiteRank
#
#      Usage
#     -----------------
#     >>> model = MultipartiteRank()
#     >>> sample =
#     'We propose a novel unsupervised keyphrase extraction approach that filters candidate keywords using outlier '
#     'detection. It starts by training word embeddings on the target document to capture semantic regularities among '
#     'the words. It then uses the minimum covariance determinant estimator to model the distribution of non-keyphrase '
#     'word vectors, under the assumption that these vectors come from the same distribution, indicative of their '
#     'irrelevance to the semantics expressed by the dimensions of the learned vector representation. Candidate '
#     'keyphrases only consist of words that are detected as outliers of this dominant distribution. Empirical results '
#     'show that our approach outperforms stateof-the-art and recent unsupervised keyphrase extraction methods.'
#     >>> model.get_keywords(sample)
#     [[
#         ('novel unsupervis keyphras extract approach', 0.2570445598381217,
#            {'stemmed': 'novel unsupervis keyphras extract approach',
#             'pos': 'ADJ ADJ NOUN NOUN NOUN',
#             'raw': ['novel unsupervised keyphrase extraction approach'],
#             'lemma': ['novel unsupervised keyphrase extraction approach'],
#             'offset': [[3, 7]],
#             'count': 1}),
#         ('keyphras word vector', 0.19388916977478182,
#            {'stemmed': 'keyphras word vector',
#             'pos': 'NOUN NOUN NOUN',
#             'raw': ['keyphrase word vectors'],
#             'lemma': ['keyphrase word vector'],
#             'offset': [[49, 51]],
#             'count': 1})
#     ]]
#     """
#
#     def __init__(self, parameter_adjust_weight: float = 1.1, *args, **kwargs):
#         """ MultipartiteRank """
#
#         super(MultipartiteRank, self).__init__(*args, **kwargs)
#         self.parameter_adjust_weight = parameter_adjust_weight
#         self.weighted_graph = True
#
#     def build_graph(self, document: str):
#         """ Build basic graph with Topic """
#
#         # convert phrase instance
#         phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
#         if len(phrase_instance) < 2:
#             # at least 2 phrase are needed to extract keywords
#             return None
#
#         # group phrases as topic classes
#         grouped_phrases = self.topic_clustering(list(phrase_instance.keys()))
#         group_id = list(range(len(grouped_phrases)))
#
#         # initialize graph instance
#         graph = nx.DiGraph()
#
#         # add nodes
#         unique_tokens_in_candidate = list(set(chain(*[i.split() for i in phrase_instance.keys()])))
#         graph.add_nodes_from(unique_tokens_in_candidate)
#
#         def get_group_id(__phrase):
#             return np.argmax([int(__phrase in g) for g in grouped_phrases])
#
#         # add edges
#         for position, __start_node in enumerate(stemmed_tokens):
#
#             # ignore invalid token
#             if __start_node not in unique_tokens_in_candidate:
#                 continue
#
#             for __position in range(position, len(stemmed_tokens)):
#                 __end_node = stemmed_tokens[__position]
#
#                 # ignore invalid token
#                 if __end_node not in unique_tokens_in_candidate:
#                     continue
#
#                 # ignore intra-topic connection
#                 if get_group_id(__start_node) == get_group_id(__end_node):
#                     continue
#
#                 weight = 1 / abs(position - __position)
#
#                 if not graph.has_edge(__start_node, __end_node):
#                     graph.add_edge(__start_node, __end_node, weight=weight)
#                     graph.add_edge(__end_node, __start_node, weight=weight)
#                 else:
#                     graph[__start_node][__end_node]['weight'] += weight
#                     graph[__end_node][__start_node]['weight'] += weight
#
#         # weight adjustment
#         for i in group_id:
#             stemmed_phrases = grouped_phrases[i]
#             offset_phrase = [(phrase_instance[phrase]['offset'][0][0], phrase) for phrase in stemmed_phrases]
#             offset_phrase_sort = sorted(offset_phrase, key=lambda key_value: key_value[0])
#             first_position, first_phrase = offset_phrase_sort[0]
#             scale = self.parameter_adjust_weight * np.exp(1 / (1+first_position))
#
#             neighbour_w = [
#                 (
#                     np.sum([
#                         graph[__start][__end]['weight']
#                         for __start, __end in graph.edges(start)
#                         if get_group_id(__start) == i and __start != first_phrase and __end == start
#                     ]),
#                     start
#                 )
#                 for start, end in graph.edges(first_phrase) if end == first_phrase
#             ]
#             for w, start in neighbour_w:
#                 graph[start][first_phrase]['weight'] += scale*w
#
#         return graph, phrase_instance, grouped_phrases, len(stemmed_tokens)
#
#     def get_keywords(self, document: str, n_keywords: int = 10):
#         """ Get keywords
#
#         Parameter
#         ------------------
#         document: str
#         n_keywords: int
#
#          Return
#         ------------------
#         a list of keyphrase with score eg) [('aa', 0.5), ('b', 0.3), ..]
#         """
#         # make graph and get data structure for candidate phrase
#         output = self.build_graph(document)
#         if output is None:
#             return []
#
#         graph, phrase_instance, grouped_phrases, original_sentence_token_size = output
#
#         # pagerank to get score for individual word (sum of score will be one)
#         node_score = self.run_pagerank(graph)
#
#         # combine score to get score of phrase
#         phrase_score = [
#             (candidate_phrase_stemmed_form, sum(node_score[t] for t in candidate_phrase_stemmed_form.split()))
#             for candidate_phrase_stemmed_form in phrase_instance.keys()
#         ]
#
#         # sorting
#         phrase_score_sorted_list = sorted(phrase_score, key=lambda key_value: key_value[1], reverse=True)
#         count_valid = min(n_keywords, len(phrase_score_sorted_list))
#
#         def modify_output(stem, score):
#             tmp = phrase_instance[stem]
#             tmp['score'] = score
#             tmp['n_source_tokens'] = original_sentence_token_size
#             return tmp
#
#         val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
#         return val



""" TFIDF based keyword extraction and wrapper of gensim API for TFIDF """
import os
import logging

import gensim
import numpy as np
from gensim import corpora

from ._phrase_constructor import PhraseConstructor

logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
CACHE_DIR = './cache/priors/tfidf'
__all__ = 'TFIDF'


class TFIDF:
    """ TFIDF """

    def __init__(self, language: str = 'en'):
        self.__model = None
        self.__dict = None
        self.prior_required = True
        self.phrase_constructor = PhraseConstructor(language=language)

    def load(self, directory: str = './cache/priors/tfidf'):
        """ load saved lda model and dictionary instance used to train the model """
        path_to_model = os.path.join(directory, 'tfidf_model')
        path_to_dict = os.path.join(directory, 'tfidf_dict')

        assert os.path.exists(path_to_model), 'no model found: {}'.format(path_to_model)
        assert os.path.exists(path_to_dict), 'no dict found: {}'.format(path_to_dict)

        logging.debug('loading model...')
        self.__model = gensim.models.TfidfModel.load(path_to_model)
        self.__dict = gensim.corpora.Dictionary.load_from_text(path_to_dict)
        self.__dict.id2token = dict([(v, k) for k, v in self.__dict.token2id.items()])

    def save(self, directory: str = './cache/priors/tfidf'):
        assert self.is_trained, 'training before run any inference'
        os.makedirs(os.path.join(directory), exist_ok=True)
        self.__model.save(os.path.join(directory, 'tfidf_model'))
        self.__dict.save_as_text(os.path.join(directory, 'tfidf_dict'))

    def train(self, data: list, export_directory: str = None, normalize: bool = True):
        """ TFIDF training

         Parameter
        ----------------
        data: list of document (list of string)
        """
        # get stemmed token
        stemmed_tokens = list(map(lambda x: self.phrase_constructor.tokenize_and_stem(x), data))
        # build TFIDF
        logging.debug("building corpus...")
        self.__dict = corpora.Dictionary(stemmed_tokens)
        self.__dict.id2token = dict([(v, k) for k, v in self.__dict.token2id.items()])
        corpus = [self.__dict.doc2bow(text) for text in stemmed_tokens]
        logging.debug("training model...")
        self.__model = gensim.models.TfidfModel(corpus=corpus, normalize=normalize)
        logging.debug("saving model and corpus at {}".format(export_directory))
        self.save(export_directory)

    def distribution_word(self, document: str):
        """ word distribution of given document with pre-calculate TFIDF matrix

         Parameter
        ----------------
        tokens: document to get topic distribution

         Return
        -----------
        dict((word_0, prob_0), ..., (word_n, prob_n)) in order of probability
        Note that `n` is dynamically changing based on the coverage of probability and the probability itself will
        change slightly due to the randomness of sampling
        """
        assert self.is_trained, 'training before run any inference'
        # get stemmed token
        stemmed_tokens = self.phrase_constructor.tokenize_and_stem(document)
        bow = self.__dict.doc2bow(stemmed_tokens)
        dist = dict((self.__dict.id2token[w_id], p) for w_id, p in self.__model[bow])
        return dist

    @property
    def model(self):
        """ gensim model instance of LDA """
        assert self.is_trained, 'training before run any inference'
        return self.__model

    @property
    def vocab_size(self):
        """ vocabulary size """
        assert self.is_trained, 'training before run any inference'
        return len(self.__dict.token2id)

    @property
    def vocab(self):
        """ vocabulary size """
        assert self.is_trained, 'training before run any inference'
        return list(self.__dict.token2id.keys())

    @property
    def dictionary(self):
        """ gensim dictionary instance """
        assert self.is_trained, 'training before run any inference'
        return self.__dict

    @property
    def is_trained(self):
        return self.__dict is not None and self.__model is not None

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of keywords with score eg) [('aa', 0.5), ('b', 0.3), ..]
        """
        # convert phrase instance
        phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
        if len(phrase_instance) < 2:
            # at least 2 phrase are needed to extract keyphrase
            return []

        dist_word = self.distribution_word(document)

        def aggregate_prob(__phrase_key):
            __phrase = phrase_instance[__phrase_key]
            prob = float(np.mean([
                dist_word[_r] if _r in dist_word.keys() else 0 for _r in __phrase['stemmed'].split()]))
            return prob

        phrase_prob = [(k, aggregate_prob(k)) for k in phrase_instance.keys()]

        # sorting
        phrase_score_sorted_list = sorted(phrase_prob, key=lambda id_score: id_score[1], reverse=True)
        count_valid = min(n_keywords, len(phrase_score_sorted_list))

        def modify_output(stem, score):
            tmp = phrase_instance[stem]
            tmp['score'] = score
            tmp['n_source_tokens'] = len(stemmed_tokens)
            return tmp

        val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
        return val



""" LexSpec: Lexical Specificity based keyword extraction """
import os
import logging
import math
import json
from math import log, factorial, log10
from collections import Counter
from itertools import chain

from ._phrase_constructor import PhraseConstructor

logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
__all__ = ('LexSpec', 'lexical_specificity', 'TF')


def average(_list):
    return sum(_list)/len(_list)


def lexical_specificity(T, t, f, k, lim: int = 400):
    """ calculate lexical specificity, originally implemented by Jose

    :param T: a size of reference corpus
    :param t: a size of sub-corpus (t <= T)
    :param f: a frequency of w in reference corpus
    :param k: a frequency of w in sub-corpus  (f <= k)
    :param lim:
    :return: float score
    """

    def stirling(n):
        return n * log(n) - n + (0.5 * log(2 * 3.141592 * n))

    assert t <= T, "t exceeds T"
    assert k <= f, "f ({}) exceeds k ({})".format(f, k)

    if f > lim:
        arg1 = stirling(f)
    else:
        arg1 = log(factorial(int(f)))
    if T - f > lim:
        arg2 = stirling(T - f)
    else:
        arg2 = log(factorial(int(T - f)))
    if t > lim:
        arg3 = stirling(t)
    else:
        arg3 = log(factorial(int(t)))
    if T - t > lim:
        arg4 = stirling(T - t)
    else:
        arg4 = log(factorial(int(T - t)))
    if T > lim:
        arg5 = stirling(T)
    else:
        arg5 = log(factorial(int(T)))
    if k > lim:
        arg6 = stirling(k)
    else:
        arg6 = log(factorial(int(k)))
    if f - k > lim:
        arg7 = stirling(f - k)
    else:
        arg7 = log(factorial(int(f - k)))
    if t - k > lim:
        arg8 = stirling(t - k)
    else:
        arg8 = log(factorial(int(t - k)))
    if T - f - t + k > lim:
        arg9 = stirling(T - f - t + k)
    else:
        arg9 = log(factorial(int(T - f - t + k)))

    prob = arg1 + arg2 + arg3 + arg4 - arg5 - arg6 - arg7 - arg8 - arg9
    first_prod = -log10(math.e) * prob

    if prob < log(0.1):
        prob1 = 1.0
        prob = 1.0
        while prob1 != 0.0 and (prob / prob1) < 10000000 and k <= f:
            prob2 = (prob1 * (f - k) * (t - k)) / ((k + 1) * (T - f - t + k + 1))
            prob = prob + prob2
            prob1 = prob2
            k += 1
        score = first_prod - log10(prob)
        return score
    else:
        return 0


class LexSpec:
    """ Lexical Specification based keyword extraction algorithm """

    def __init__(self, language: str = 'en'):
        self.__reference_corpus = None
        self.freq = None
        self.reference_corpus_size = 0
        self.prior_required = True
        self.phrase_constructor = PhraseConstructor(language=language)

    def load(self, directory: str = './cache/priors/lexical_specificity'):
        """ load saved lda model and dictionary instance used to train the model """
        path_to_dict = "{}/lexical_specificity_frequency.json".format(directory)
        assert os.path.exists(path_to_dict), 'no dictionary found: {}'.format(path_to_dict)
        with open(path_to_dict, 'r') as f:
            self.freq = json.load(f)
        self.reference_corpus_size = sum(self.freq.values())
        logging.debug('loaded frequency dictionary from {}'.format(path_to_dict))

    def train(self, data: list, export_directory: str = './cache/priors/lexical_specificity'):
        """ cache a dictionary {key: a word, value: occcurence of the word}

         Parameter
        ----------------
        data: a list of document (list of string)
        """
        # get stemmed token
        stemmed_tokens = list(chain(*map(lambda x: self.phrase_constructor.tokenize_and_stem(x), data)))
        self.freq = dict(Counter(stemmed_tokens))
        self.reference_corpus_size = sum(self.freq.values())
        self.save(export_directory)
        logging.debug('compute frequency dictionary saved at {}'.format(export_directory))

    def save(self, directory: str = './cache/priors/lexical_specificity'):
        assert self.is_trained, 'training before run any inference'
        os.makedirs(os.path.join(directory), exist_ok=True)
        with open("{}/lexical_specificity_frequency.json".format(directory), "w") as f:
            json.dump(self.freq, f)

    @property
    def is_trained(self):
        return self.freq is not None

    def lexical_specificity(self, document: str):
        stemmed_tokens = self.phrase_constructor.tokenize_and_stem(document)
        sub_freq = dict(Counter(stemmed_tokens))

        # compute lexical specificity
        sub_corpus_size = sum(sub_freq.values())
        ls_dict = {k: lexical_specificity(
            T=self.reference_corpus_size,
            t=sub_corpus_size,
            f=self.freq[k],
            k=sub_freq[k]
        ) for k in sub_freq.keys() if k in self.freq.keys()}
        return ls_dict

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of dictionary consisting of 'stemmed', 'pos', 'raw', 'offset', 'count'.
        eg) {'stemmed': 'grid comput', 'pos': 'ADJ NOUN', 'raw': ['grid computing'], 'offset': [[11, 12]], 'count': 1}
        """
        assert self.is_trained, 'provide prior before running inference'

        # convert phrase instance
        phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
        if len(phrase_instance) < 2:
            # at least 2 phrase are needed to extract keyphrase
            return []

        sub_freq = dict(Counter(stemmed_tokens))

        # compute lexical specificity
        sub_corpus_size = sum(sub_freq.values())
        ls_dict = {k: lexical_specificity(
            T=self.reference_corpus_size,
            t=sub_corpus_size,
            f=self.freq[k],
            k=sub_freq[k]
        ) for k in sub_freq.keys() if k in self.freq.keys()}

        # aggregate score over individual word in a phrase
        def aggregate_prob(__phrase_key):
            __phrase = phrase_instance[__phrase_key]
            prob = float(average([
                ls_dict[_r] if _r in ls_dict.keys() else 0 for _r in __phrase['stemmed'].split()]))
            return prob

        phrase_prob = [(k, aggregate_prob(k)) for k in phrase_instance.keys()]

        # sorting
        phrase_score_sorted_list = sorted(phrase_prob, key=lambda id_score: id_score[1], reverse=True)
        count_valid = min(n_keywords, len(phrase_score_sorted_list))

        def modify_output(stem, score):
            tmp = phrase_instance[stem]
            tmp['score'] = score
            tmp['n_source_tokens'] = len(stemmed_tokens)
            return tmp

        val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
        return val


class TF(LexSpec):
    """ Term frequency based keyword extraction algorithm """

    def __init__(self, *args, **kwargs):
        """ Term frequency based keyword extraction algorithm """
        super(TF, self).__init__(*args, **kwargs)

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of dictionary consisting of 'stemmed', 'pos', 'raw', 'offset', 'count'.
        eg) {'stemmed': 'grid comput', 'pos': 'ADJ NOUN', 'raw': ['grid computing'], 'offset': [[11, 12]], 'count': 1}
        """
        assert self.is_trained, 'provide prior before running inference'

        # convert phrase instance
        phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
        if len(phrase_instance) < 2:
            # at least 2 phrase are needed to extract keyphrase
            return []

        sub_freq = dict(Counter(stemmed_tokens))

        # compute term frequency
        tf_dict = {k: self.freq[k] for k in sub_freq.keys() if k in self.freq.keys()}

        # aggregate score over individual word in a phrase
        def aggregate_prob(__phrase_key):
            __phrase = phrase_instance[__phrase_key]
            prob = float(average([
                tf_dict[_r] if _r in tf_dict.keys() else 0 for _r in __phrase['stemmed'].split()]))
            return prob

        phrase_prob = [(k, aggregate_prob(k)) for k in phrase_instance.keys()]

        # sorting
        phrase_score_sorted_list = sorted(phrase_prob, key=lambda id_score: id_score[1], reverse=True)
        count_valid = min(n_keywords, len(phrase_score_sorted_list))

        def modify_output(stem, score):
            tmp = phrase_instance[stem]
            tmp['score'] = score
            tmp['n_source_tokens'] = len(stemmed_tokens)
            return tmp

        val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
        return val




import os
import re
from glob import glob

from ._phrase_constructor import PhraseConstructor
from nltk.stem.porter import PorterStemmer  # only for English

STEMMER = PorterStemmer()
VALID_DATASET_LIST = [
    # "110-PT-BN-KP", "cacic", "pak2018", "WikiNews", "wicc"
    "citeulike180", "fao30", "fao780", "Inspec", "kdd", "Krapivin2009", "Nguyen2007", "PubMed", "Schutz2008",
    "SemEval2010", "SemEval2017", "theses100", "wiki20", "www", "500N-KPCrowd-v1.1"
]

__all__ = ('get_benchmark_dataset', "VALID_DATASET_LIST")


def preprocess_source(string):
    escaped_punctuation = {'-lrb-': '(', '-rrb-': ')', '-lsb-': '[', '-rsb-': ']', '-lcb-': '{', '-rcb-': '}'}
    for k, v in escaped_punctuation.items():
        string = string.replace(k, v)
    return string


def split_for_keywords(string):

    def cleaner(_string):
        _string = re.sub(r'\A\s*', '', _string)
        _string = re.sub(r'\s*\Z', '', _string)
        _string = _string.replace('\n', ' ').replace('\t', '')
        _string = re.sub(r'\s{2,}', ' ', _string)
        _string = ' '.join(list(map(lambda x: STEMMER.stem(x), _string.split(' '))))
        return _string

    keys = list(filter(
        lambda x: len(x) > 0, [cleaner(s) for s in re.split('[\n,]', string)])
    )
    return list(set(keys))


def get_benchmark_dataset(data: str = 'Inspec',
                          cache_dir: str = "./cache",
                          keep_only_valid_label: bool = True):
    """ Get a dataset for keyword extraction (all not stemmed)

     Parameter
    -------------
    data: str
        dataset name
    cache_dir: str
        directory to cache the data
    keep_only_valid_label: bool
        False to get all the label, otherwise only keep the label set that is in the phrase candidates

     Return
    -------------
    dictionary consists of the data
    language
    """
    assert data in VALID_DATASET_LIST, "undefined dataset: {}".format(data)
    url = "https://github.com/LIAAD/KeywordExtractor-Datasets/raw/master/datasets"
    os.makedirs(cache_dir, exist_ok=True)
    if not os.path.exists("{}/{}".format(cache_dir, data)):
        os.system('wget -O {0}/{2}.zip {1}/{2}.zip'.format(cache_dir, url, data))
        if data == "110-PT-BN-KP":
            os.system("unzip {0}/{1}.zip -d {0}/{1}".format(cache_dir, data))
        elif data == "WikiNews":
            os.system("unzip {0}/{1}.zip -d {0}".format(cache_dir, data))
            os.system("mv {0}/WKC {0}/{1}".format(cache_dir, data))
        else:
            os.system("unzip {0}/{1}.zip -d {0}/".format(cache_dir, data))
    if not os.path.exists("{}/{}".format(cache_dir, data)):
        raise ValueError('data `{}` is not in the list {}'.format(data, url))

    def safe_open(_file):
        with open(_file, 'r') as f:
            return f.read()

    language = safe_open("{}/{}/language.txt".format(cache_dir, data)).replace('\n', '')
    answer_dict = [{
        "keywords":
            split_for_keywords(safe_open(
                "{}/{}/keys/{}.key".format(cache_dir, data, '.txt'.join(os.path.basename(t).split('.txt')[:-1])))),
        "source": preprocess_source(safe_open(t)),
        "id": os.path.basename(t)} for t in glob("{}/{}/docsutf8/*.txt".format(cache_dir, data))]

    if not keep_only_valid_label:
        return answer_dict, language

    phraser = PhraseConstructor()

    def _filter_label(_dict):
        phrase, stemmed_token = phraser.tokenize_and_stem_and_phrase(_dict['source'])
        keywords_valid = list(set(phrase.keys()).intersection(set(_dict['keywords'])))
        if len(keywords_valid) == 0:
            return None
        else:
            _dict['keywords'] = keywords_valid
            return _dict

    answer_dict = list(map(lambda x: _filter_label(x), answer_dict))
    answer_dict = list(filter(None, answer_dict))
    return answer_dict, language




import kex


VALID_ALGORITHMS = [
    'FirstN', 'TF', 'LexSpec', 'TFIDF', 'TextRank', 'SingleRank', 'PositionRank', 'LexRank', 'TFIDFRank', 'SingleTPR',
    'TopicRank'
]


def get_algorithm(model_name, *args, **kwargs):

    if model_name == 'TopicRank':
        model = kex.TopicRank(*args, **kwargs)
    elif model_name == 'TextRank':
        model = kex.TextRank(*args, **kwargs)
    elif model_name == 'SingleRank':
        model = kex.SingleRank(*args, **kwargs)
    elif model_name == 'TFIDFRank':
        model = kex.TFIDFRank(*args, **kwargs)
    elif model_name == 'PositionRank':
        model = kex.PositionRank(*args, **kwargs)
    elif model_name == 'TFIDF':
        model = kex.TFIDF(*args, **kwargs)
    elif model_name == 'LexSpec':
        model = kex.LexSpec(*args, **kwargs)
    elif model_name == 'LexRank':
        model = kex.LexRank(*args, **kwargs)
    elif model_name == 'FirstN':
        model = kex.FirstN(*args, **kwargs)
    elif model_name == 'TF':
        model = kex.TF(*args, **kwargs)
    elif model_name == 'TopicalPageRank':
        model = kex.TopicalPageRank(*args, **kwargs)
    elif model_name == 'SingleTPR':
        model = kex.SingleTPR(*args, **kwargs)
    else:
        raise ValueError('unknown model: {}\n valid algorithms are: {}'.format(model_name, VALID_ALGORITHMS))

    return model



""" Implementation of the TopicalPageRank """
import numpy as np
from .lda import LDA
from .graph_text_rank import TextRank


class TopicalPageRank(TextRank):
    """ TopicalPageRank """

    def __init__(self, language: str = 'en', *args, **kwargs):
        """ TopicalPageRank """

        super(TopicalPageRank, self).__init__(language=language, *args, **kwargs)
        self.lda = LDA(language=language)
        self.directed_graph = True
        self.weighted_graph = True
        self.prior_required = True

    def load(self, directory: str = None):
        self.lda.load(directory)

    def train(self, data: list, export_directory: str = None, num_topics: int = 15):
        self.lda.train(data, export_directory=export_directory, num_topics=num_topics)

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of dictionary consisting of 'stemmed', 'pos', 'raw', 'offset', 'count'.
        eg) {'stemmed': 'grid comput', 'pos': 'ADJ NOUN', 'raw': ['grid computing'], 'offset': [[11, 12]], 'count': 1}
        """
        # make graph and get data structure for candidate phrase
        output = self.build_graph(document)
        if output is None:
            return []

        # make graph and get data structure for candidate phrase
        graph, phrase_instance, original_sentence_token_size = self.build_graph(document)

        # pagerank to get score for individual word (sum of score will be one)
        word_matrix = self.lda.probability_word()  # topic size x vocab
        topic_dist = self.lda.distribution_topic_document(document)
        topic_vector = np.array([i for t, i in topic_dist])  # topic size

        # original TPR procedure
        __node_score = []

        for i in range(self.lda.topic_size):
            bias = dict()
            vocab = self.lda.dictionary.token2id
            unk = 0
            for word in graph.nodes():
                if word in vocab.keys():
                    word_id = vocab[word]
                    bias[word] = word_matrix[i, word_id]
                else:
                    bias[word] = 0.0
                    unk += 1

            # normalize the topical word importance of words
            norm = sum(bias.values())
            for word in bias:
                if norm == 0:
                    bias[word] = 1 / len(bias[word])
                else:
                    bias[word] /= norm
            __node_score.append(self.run_pagerank(graph, personalization=bias))

        # combine score to get score of phrase
        phrase_score_dict = [
            (candidate_phrase_stemmed_form, sum(
                sum(__node_score[i][t] for t in candidate_phrase_stemmed_form.split()) * topic_vector[i]
                for i in range(self.lda.topic_size)))
            for candidate_phrase_stemmed_form in phrase_instance.keys()
        ]

        # sorting
        phrase_score_sorted_list = sorted(phrase_score_dict, key=lambda key_value: key_value[1], reverse=True)
        count_valid = min(n_keywords, len(phrase_score_sorted_list))

        def modify_output(stem, _score):
            tmp = phrase_instance[stem]
            tmp['score'] = _score
            tmp['n_source_tokens'] = original_sentence_token_size
            return tmp

        val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
        return val



""" Implementation of basic TextRank variants
 * TextRank: Undirected Unweighted Graph
 * SingleRank: Undirected Weighted Graph (weight is co-occurrence)
 * PositionRank: Undirected Weighted Graph (weight is co-occurrence, bias is offset)
 * TFIDFRank: Undirected Weighted Graph (weight is co-occurrence, bias is TFIDF word probability)
 * LexRank: Undirected Weighted Graph (weight is co-occurrence, bias is lexical specificity)
 * SingleTPR: Directed Weighted Graph (weight is co-occurrence, bias is LDA topic x word distribution)
"""
import numpy as np
import networkx as nx
from itertools import chain

from ._phrase_constructor import PhraseConstructor
from .tfidf import TFIDF
from .lda import LDA
from .lexical_specificity import LexSpec

__all__ = ('TextRank', 'SingleRank', 'PositionRank', 'TFIDFRank', 'SingleTPR', 'LexRank')


class TextRank:
    """ TextRank """

    def __init__(self,
                 language: str = 'en',
                 window_size: int = 2,
                 random_prob: float = 0.85,
                 tol: float = 0.0001):
        """ TextRank

         Parameter
        ------------------------
        language: str
            `en` or `ja`
        window_size: int
            window size to make graph edges
        random_prob: float
            PageRank parameter that is coefficient of convex combination of random suffer model
        tol: float
            PageRank parameter that define tolerance of convergence
        """
        self.__window_size = window_size
        self.__random_prob = random_prob
        self.__tol = tol

        self.phrase_constructor = PhraseConstructor(language=language)
        self.weighted_graph = False
        self.direct_graph = False
        self.prior_required = False
        self.normalize = True
        self.page_rank_bias_type = None
        self.tfidf = None
        self.lexical = None
        self.lda = None

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of dictionary consisting of 'stemmed', 'pos', 'raw', 'offset', 'count'.
        eg) {'stemmed': 'grid comput', 'pos': 'ADJ NOUN', 'raw': ['grid computing'], 'offset': [[11, 12]], 'count': 1}
        """
        # make graph and get data structure for candidate phrase
        output = self.build_graph(document)
        if output is None:
            return []

        graph, phrase_instance, original_sentence_token_size, bias = output
        node_score = self.run_pagerank(graph, personalization=bias)

        # combine score to get score of phrase
        phrase_score = [
            (
                candidate_phrase_stemmed_form,
                sum(node_score[t] for t in candidate_phrase_stemmed_form.split())
                / len(candidate_phrase_stemmed_form.split())  # - v['offset'][0][0] * 1e-8
                if self.normalize else sum(node_score[t] for t in candidate_phrase_stemmed_form.split())
            ) for candidate_phrase_stemmed_form, v in phrase_instance.items()
        ]
        # sorting
        phrase_score_sorted_list = sorted(phrase_score, key=lambda key_value: key_value[1], reverse=True)
        count_valid = min(n_keywords, len(phrase_score_sorted_list))

        def modify_output(stem, score):
            tmp = phrase_instance[stem]
            tmp['score'] = score
            tmp['n_source_tokens'] = original_sentence_token_size
            return tmp

        val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
        return val

    def build_graph(self, document: str):
        """ Build basic graph
        - nodes: phrases extracted from given document
        - edge: co-occurrence in certain window
        - weight: count of co-occurrence

         Parameter
        ------------------------
        document: str
            single document

         Return
        ------------------------
        graph:
            graph instance
        phrase_instance:
            phrase instance
        original_sentence_token_size: int
            token size in target_document
        """

        # convert phrase instance
        phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
        if len(phrase_instance) < 2:
            # at least 2 phrase are needed to extract keyphrase
            return None

        # initialize graph instance
        if self.direct_graph:
            graph = nx.DiGraph()
        else:
            graph = nx.Graph()

        # add nodes
        unique_tokens_in_candidate = list(set(chain(*[i.split() for i in phrase_instance.keys()])))
        graph.add_nodes_from(unique_tokens_in_candidate)

        # add edges
        for position, __start_node in enumerate(stemmed_tokens):

            # ignore invalid token
            if __start_node not in unique_tokens_in_candidate:
                continue

            for __position in range(position, min(position + self.__window_size, len(stemmed_tokens))):
                __end_node = stemmed_tokens[__position]

                # ignore invalid token
                if __end_node not in unique_tokens_in_candidate:
                    continue

                if __start_node == __end_node:
                    continue

                if not graph.has_edge(__start_node, __end_node):
                    if self.weighted_graph:
                        graph.add_edge(__start_node, __end_node, weight=0.0)
                    else:
                        graph.add_edge(__start_node, __end_node)

                if self.weighted_graph:
                    # SingleRank employ weight as the co-occurrence times
                    graph[__start_node][__end_node]['weight'] += 1.0
        if self.page_rank_bias_type == 'position_rank':
            bias = {}
            for n, s in enumerate(stemmed_tokens):
                if s not in unique_tokens_in_candidate:
                    continue
                if s not in bias.keys():
                    bias[s] = 1 / (1 + n)
                else:
                    bias[s] += 1 / (1 + n)
            norm = sum(bias.values())
            bias = {k: v/norm for k, v in bias.items()}
        elif self.page_rank_bias_type == 'lda':
            # pagerank to get score for individual word (sum of score will be one)
            word_matrix = self.lda.probability_word()  # topic size x vocab
            topic_dist = self.lda.distribution_topic_document(document)
            topic_vector = np.array([i for t, i in topic_dist])  # topic size

            # single TPR
            vocab = self.lda.dictionary.token2id

            def norm_inner(a, b):
                return np.sum(a * b) / (np.sqrt(np.sum(a * a) + np.sum(b * b)) + 1e-7)

            bias = dict()
            for word in unique_tokens_in_candidate:
                if word in vocab.keys():
                    word_id = vocab[word]
                    word_vector = word_matrix[:, word_id]
                    bias[word] = norm_inner(word_vector, topic_vector)
                else:
                    bias[word] = 0.0

            # normalize the topical word importance of words
            norm = sum(bias.values())
            bias = {k: v / norm for k, v in bias.items()}
        elif self.page_rank_bias_type in ['tfidf', 'lex_spec']:
            if self.page_rank_bias_type == 'tfidf':
                dist_word = self.tfidf.distribution_word(document)
            else:
                dist_word = self.lexical.lexical_specificity(document)
            bias = {word: dist_word[word] if word in dist_word.keys() else 0.0 for word in unique_tokens_in_candidate}
            norm = sum(bias.values())
            bias = {k: v / norm for k, v in bias.items()}
        else:
            bias = None
        return graph, phrase_instance, len(stemmed_tokens), bias

    def run_pagerank(self, graph, personalization=None):
        if personalization:
            return nx.pagerank(
                G=graph, alpha=self.__random_prob, tol=self.__tol, weight='weight', personalization=personalization)
        else:
            return nx.pagerank_scipy(G=graph, alpha=self.__random_prob, tol=self.__tol, weight='weight')


class SingleRank(TextRank):
    """ SingleRank """

    def __init__(self, window_size: int = 2, *args, **kwargs):
        """ SingleRank """
        super(SingleRank, self).__init__(window_size=window_size, *args, **kwargs)
        self.weighted_graph = True


class PositionRank(TextRank):
    """ PositionRank """

    def __init__(self, window_size: int = 2, *args, **kwargs):
        """ PositionRank """
        super(PositionRank, self).__init__(window_size=window_size, *args, **kwargs)
        self.weighted_graph = True
        self.page_rank_bias_type = 'position_rank'


class TFIDFRank(TextRank):
    """ TFIDFRank """

    def __init__(self, language: str = 'en', *args, **kwargs):
        """ ExpandRank """
        super(TFIDFRank, self).__init__(language=language, *args, **kwargs)
        self.tfidf = TFIDF(language=language)
        self.weighted_graph = True
        self.prior_required = True
        self.page_rank_bias_type = 'tfidf'

    def load(self, directory: str = None):
        self.tfidf.load(directory)

    def train(self, data: list, export_directory: str = None):
        self.tfidf.train(data, export_directory=export_directory)


class LexRank(TextRank):
    """ LexRank """

    def __init__(self, language: str = 'en', *args, **kwargs):
        """ LexRank """
        super(LexRank, self).__init__(language=language, *args, **kwargs)
        self.lexical = LexSpec(language=language)
        self.weighted_graph = True
        self.prior_required = True
        self.page_rank_bias_type = 'lex_spec'

    def load(self, directory: str = None):
        self.lexical.load(directory)

    def train(self, data: list, export_directory: str = None):
        self.lexical.train(data, export_directory=export_directory)


class SingleTPR(TextRank):
    """ Single TopicalPageRank """

    def __init__(self, language: str = 'en', *args, **kwargs):
        """ Single TopicalPageRank """

        super(SingleTPR, self).__init__(language=language, *args, **kwargs)
        self.lda = LDA(language=language)
        self.directed_graph = True
        self.weighted_graph = True
        self.prior_required = True
        self.page_rank_bias_type = 'lda'

    def load(self, directory: str = None):
        self.lda.load(directory)

    def train(self, data: list, export_directory: str = None, num_topics: int = 15):
        self.lda.train(data, export_directory=export_directory, num_topics=num_topics)



""" Implementation of heuristic methods """
from ._phrase_constructor import PhraseConstructor

__all__ = 'FirstN'


class FirstN:
    """ First N keywords extractor """

    def __init__(self, language: str = 'en', maximum_word_number: int = 3):
        """ First N keywords extractor """
        self.phrase_constructor = PhraseConstructor(language=language, maximum_word_number=maximum_word_number)
        self.prior_required = False

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of dictionary consisting of 'stemmed', 'pos', 'raw', 'offset', 'count'.
        eg) {'stemmed': 'grid comput', 'pos': 'ADJ NOUN', 'raw': ['grid computing'], 'offset': [[11, 12]], 'count': 1}
        """
        phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
        sorted_phrases = sorted(phrase_instance.values(), key=lambda x: x['offset'][0][0])
        sorted_phrases = sorted_phrases[:min(len(sorted_phrases), n_keywords)]
        for n, i in enumerate(sorted_phrases):
            sorted_phrases[n]['score'] = n

        return sorted_phrases



""" Wrapper of gensim API for a Topic Modeling with LDA """
import os
import logging

import gensim
from gensim import corpora

from ._phrase_constructor import PhraseConstructor

logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
__all__ = 'LDA'


class LDA:
    """ LDA """

    def __init__(self, language: str = 'en'):
        self.__model = None
        self.__dict = None
        self.phrase_constructor = PhraseConstructor(language=language)

    def load(self, directory: str = './cache/priors/lda'):
        """ load saved lda model and dictionary instance used to train the model """
        path_to_model = os.path.join(directory, 'lda_model')
        path_to_dict = os.path.join(directory, 'lda_dict')

        assert os.path.exists(path_to_model), 'no model found: {}'.format(path_to_model)
        assert os.path.exists(path_to_dict), 'no dict found: {}'.format(path_to_dict)

        logging.debug('loading model...')
        self.__model = gensim.models.ldamodel.LdaModel.load(path_to_model)
        self.__dict = gensim.corpora.Dictionary.load_from_text(path_to_dict)
        self.__dict.id2token = dict([(v, k) for k, v in self.__dict.token2id.items()])

    def save(self, directory: str = './cache/priors/lda'):
        assert self.is_trained, 'training before run any inference'
        os.makedirs(os.path.join(directory), exist_ok=True)
        self.__model.save(os.path.join(directory, 'lda_model'))
        self.__dict.save_as_text(os.path.join(directory, 'lda_dict'))

    def train(self, data: list, export_directory: str = None, num_topics: int = 15):
        """ LDA training

         Parameter
        ----------------
        data: list of document (list of string)
        num_topics: number of topic
        """
        stemmed_tokens = list(map(lambda x: self.phrase_constructor.tokenize_and_stem(x), data))
        # build LDA model
        logging.debug("building corpus...")
        self.__dict = corpora.Dictionary(stemmed_tokens)
        self.__dict.id2token = dict([(v, k) for k, v in self.__dict.token2id.items()])
        corpus = [self.__dict.doc2bow(text) for text in stemmed_tokens]
        logging.debug("training model...")
        self.__model = gensim.models.ldamodel.LdaModel(
            corpus=corpus, num_topics=num_topics, id2word=self.__dict)
        logging.debug("saving model and corpus at {}".format(export_directory))
        self.save(export_directory)

    def distribution_topic_document(self, document: str):
        """ topic probability distribution of given documents p(z|d = tokens)

         Parameter
        ----------------
        tokens: document to get topic distribution

         Return
        -----------
        [(topic_id, prob), (topic_id, prob), ...] in order of prob
        """
        assert self.is_trained, 'training before run any inference'
        stemmed_tokens = self.phrase_constructor.tokenize_and_stem(document)
        bow = self.__dict.doc2bow(stemmed_tokens)
        topic_dist = self.__model.get_document_topics(bow, minimum_probability=0.0)
        return topic_dist

    def probability_word(self, tokens: list = None):
        """ individual probability of given tokens conditioned by topic: [p(w_i|z) for w_i in tokens]

         Parameter
        ----------------
        tokens: list of token, each of them is treated as independently and calculate probability conditioned
                by each topic

         Return
        -----------
        ndarray (topic_number, len(tokens)): element (j, i) corresponds to a probability of token_i conditioned by
        topic j, p(w_i|z_j). If tokens is not provided, return entire probability distribution.
        """
        assert self.is_trained, 'training before run any inference'
        topic_table = self.__model.get_topics()
        if tokens is None:
            return topic_table
        else:
            ids = [self.__dict.token2id[t] for t in tokens]
            topic_dist = topic_table[:, ids]
            return topic_dist

    @property
    def is_trained(self):
        return self.__dict is not None and self.__model is not None

    @property
    def model(self):
        """ gensim model instance of LDA """
        assert self.is_trained, 'training before run any inference'
        return self.__model

    @property
    def vocab_size(self):
        """ vocabulary size """
        assert self.is_trained, 'training before run any inference'
        return len(self.__dict.token2id)

    @property
    def vocab(self):
        """ vocabulary size """
        assert self.is_trained, 'training before run any inference'
        return list(self.__dict.token2id.keys())

    @property
    def topic_size(self):
        """ topic size """
        assert self.is_trained, 'training before run any inference'
        return self.__model.num_topics

    @property
    def dictionary(self):
        """ gensim dictionary instance """
        assert self.is_trained, 'training before run any inference'
        return self.__dict



""" tokenizer/stemmer/PoS tagger/phrase constructor """
import re
from typing import List
from itertools import chain

import nltk
from segtok.segmenter import split_multi
from segtok.tokenizer import web_tokenizer, split_contractions
from nltk.stem.porter import PorterStemmer
from nltk import pos_tag

from ._stopwords import get_stopwords_list


__all__ = 'PhraseConstructor'


class Phrase:
    """ Data structure for phrases, which follows regx [Adjective]*[Noun|proper noun]+

    The structure contains a main dictionary, in which key is the stemmed candidates (so, phrases which share same
    stemmed form are regarded as same phrase)
        * pos: pos
        * stemmed: stemmed form
        * raw: list of raw phrase
        * offset: offset
    Also exclude token in given stop word list.
    """

    def __init__(self,
                 join_without_space: bool = False,
                 maximum_word_number: int = None,
                 maximum_char_number: int = None):
        """ Data structure for phrases, which follows regx [Adjective]*[Noun|proper noun]+

         Parameter
        -------------------
        join_without_space: join token without halfspace to construct a phrase (eg, Japanese)
        """

        self.__content = dict()
        self.__initialize_list()
        self.__joiner = '' if join_without_space else ' '
        self.__maximum_word_number = maximum_word_number
        self.__maximum_char_number = maximum_char_number

    @property
    def phrase(self):
        return self.__content

    def __initialize_list(self):
        self.__tmp_phrase_raw = []
        self.__tmp_phrase_stemmed = []
        self.__tmp_phrase_pos = []
        self.__tmp_phrase_offset = []

    def add(self, raw: str, stemmed: str, pos: str, offset: int):
        """ add single token, integrate it as phrase or keep as part of phrase

         Parameter
        -------------------
        raw: token in raw form
        stemmed: stemmed token
        pos: Part of speech
        offset: offset position
        """

        def add_tmp_list():  # add to tmp list
            self.__tmp_phrase_raw.append(raw)
            self.__tmp_phrase_stemmed.append(stemmed)
            self.__tmp_phrase_pos.append(pos)
            self.__tmp_phrase_offset.append(offset)

        # phrase with more symbol than alphanumeric should be ignored
        if len(re.sub(r'\w', '', stemmed)) > len(re.sub(r'\W', '', stemmed)):
            pos = 'RANDOM'

        if pos == 'NOUN' or (pos == 'ADJ' and 'NOUN' not in self.__tmp_phrase_pos):
            add_tmp_list()
        else:
            if 'NOUN' in self.__tmp_phrase_pos:  # finalize list of tokens as phrase
                self.__add_to_structure()
            else:  # only adjective can't be regarded as phrase
                self.__initialize_list()

    def __add_to_structure(self):
        """ add to main dictionary and initialize tmp lists """
        phrase_stemmed = self.__joiner.join(self.__tmp_phrase_stemmed)
        # too many words or too long character should be ignored
        if (self.__maximum_word_number and len(self.__tmp_phrase_stemmed) > self.__maximum_word_number) or \
                (self.__maximum_char_number and len(phrase_stemmed) > self.__maximum_char_number):
            self.__initialize_list()
            return

        if len(self.__tmp_phrase_offset) == 1:
            offset = [self.__tmp_phrase_offset[0], self.__tmp_phrase_offset[0] + 1]
        else:
            offset = [self.__tmp_phrase_offset[0], self.__tmp_phrase_offset[-1]]

        # key of main dictionary
        if phrase_stemmed in self.__content.keys():
            self.__content[phrase_stemmed]['raw'] += [self.__joiner.join(self.__tmp_phrase_raw)]
            self.__content[phrase_stemmed]['offset'] += [offset]
            self.__content[phrase_stemmed]['count'] += 1
        else:
            self.__content[phrase_stemmed] = dict(
                stemmed=self.__joiner.join(self.__tmp_phrase_stemmed), pos=' '.join(self.__tmp_phrase_pos),
                raw=[self.__joiner.join(self.__tmp_phrase_raw)], offset=[offset], count=1)

        # initialize tmp lists
        self.__initialize_list()
        return


class PhraseConstructor:
    """ Phrase constructor to extract a list of phrase from given sentence based on `Phrase` class """

    def __init__(self,
                 language: str = 'en',
                 stopwords_list: List = None,
                 maximum_word_number: int = 6,
                 maximum_char_number: int = 70):
        """ Phrase constructor to extract a list of phrase from given sentence based on `Phrase` class

         Parameter
        ----------------------
        language: str
        stopwords_list: List
        """
        self.__language = language.lower()[:2]
        self.__stopwords = get_stopwords_list(self.__language, stopwords_list=stopwords_list)
        self.__maximum_word_number = maximum_word_number
        self.__maximum_char_number = maximum_char_number
        if self.__language == 'en':
            self.__stemmer = PorterStemmer()
            try:
                pos_tag(['this', 'is', 'a', 'sample'])
            except LookupError:
                nltk.download('averaged_perceptron_tagger')
            self.__pos_tagger = pos_tag
        else:
            raise ValueError('available only for English because of PoS tagger')

    @staticmethod
    def simplify_pos(_pos):
        if _pos in ["NN", "NNS", "NNP", "NNPS", "CD"]:
            return "NOUN"
        elif _pos in ["JJ", "JJR", "JJS"]:
            return "ADJ"
        return "RANDOM"

    def tokenize_and_stem(self, document: str, apply_stopwords: bool = True):
        _, stemmed, _ = self.preprocess(document)
        if apply_stopwords:
            return list(filter(lambda x: x.lower() not in self.__stopwords, stemmed))
        return stemmed

    def preprocess(self, document: str, flatten_sentence: bool = True):
        """ tokenization/stemming/PoS tagging """
        sentence_token = [[
            w for w in split_contractions(web_tokenizer(s)) if not (w.startswith("'") and len(w) > 1) and len(w) > 0
        ] for s in list(split_multi(document)) if len(s.strip()) > 0]
        stemmed = [list(map(lambda x: self.__stemmer.stem(x), words)) for words in sentence_token]
        pos = [list(map(lambda x: self.simplify_pos(x[1]), self.__pos_tagger(words))) for words in sentence_token]
        if flatten_sentence:
            return list(chain(*sentence_token)), list(chain(*stemmed)), list(chain(*pos))
        return sentence_token, stemmed, pos

    def tokenize_and_stem_and_phrase(self, document: str):
        """ tokenization & stemming & phrasing

         Parameter
        ---------------------
        document: str

         Return
        ---------------------
        `Phrase.phrase` object, stemmed_token
        """
        sentence_token, stemmed, pos = self.preprocess(document)

        # phraser instance
        phrase_structure = Phrase(self.__language == 'ja',
                                  maximum_word_number=self.__maximum_word_number,
                                  maximum_char_number=self.__maximum_char_number)
        n = 0
        for n, (t, s, p) in enumerate(zip(sentence_token, stemmed, pos)):
            if t.lower() in self.__stopwords or s.lower() in self.__stopwords:
                p = 'RANDOM'
            phrase_structure.add(raw=t, stemmed=s, pos=p, offset=n)

        # to finalize the phrase in the end of sentence
        phrase_structure.add(raw='.', stemmed='.', pos='PUNCT', offset=n + 1)

        return phrase_structure.phrase, stemmed,

    def force_reset_stopwords(self):
        self.__stopwords = []

    @property
    def stopwords(self):
        return self.__stopwords



""" stopwords are fetched from https://github.com/LIAAD/yake/tree/master/yake/StopwordsList """
import os
from typing import List


def get_stopwords_list(language: str = 'en', stopwords_list: List = None):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    local_path = os.path.join("_stopwords", "stopwords_{}.txt".format(language[:2]))
    if not os.path.exists(os.path.join(dir_path, local_path)):
        local_path = os.path.join("StopwordsList", "stopwords_noLang.txt")

    resource_path = os.path.join(dir_path, local_path)
    try:
        with open(resource_path, encoding='utf-8') as stop_fil:
            _stopwords_list = stop_fil.read().lower().split("\n")
    except UnicodeEncodeError:
        with open(resource_path, encoding='ISO-8859-1') as stop_fil:
            _stopwords_list = stop_fil.read().lower().split("\n")

    if stopwords_list:
        return list(set(_stopwords_list + stopwords_list))
    else:
        return list(set(_stopwords_list))



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='kex',
    packages=find_packages(),
    version='1.0.1',
    license='MIT',
    description='Light/easy keyword extraction.',
    url='https://github.com/asahi417/kex',
    download_url='https://github.com/asahi417/kex/archive/v1.1.tar.gz',
    keywords=['keyword-extraction', 'nlp', 'information-retrieval'],
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Asahi Ushio',
    author_email='asahi1992ushio@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',       # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
      ],
    include_package_data=True,
    test_suite='tests',
    install_requires=[
        'gensim>=3.4.0,<3.5.0',
        'nltk==3.5',
        'networkx',
        'numpy>=1.16.1',
        'segtok'
    ],
    python_requires='>=3.6',
)



""" Implementation of TopicRank """
from itertools import chain, combinations, product

import networkx as nx
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import pdist
from ._phrase_constructor import PhraseConstructor

__all__ = 'TopicRank'


class TopicRank:
    """ TopicRank """

    def __init__(self,
                 language: str = 'en',
                 random_prob: float = 0.85,
                 tol: float = 0.0001,
                 clustering_threshold: float = 0.74,
                 linkage_method: str = 'average'):
        """ TopicRank

         Parameter
        ------------------------
        language: str
            `en` or `ja`
        window_size: int
            window size to make graph edges
        random_prob: float
            PageRank parameter that is coefficient of convex combination of random suffer model
        tol: float
            PageRank parameter that define tolerance of convergence
        clustering_threshold: float
            threshold for clustering algorithm
        linkage_method: str
            graph linkage method
        add_verb: bool
            take verbs into account
        no_stemming: bool
            no stemming is applied (True if the document is already stemmed)
        """
        self.prior_required = False
        self.__random_prob = random_prob
        self.__tol = tol

        self.__linkage_method = linkage_method
        self.__clustering_threshold = clustering_threshold

        self.phrase_constructor = PhraseConstructor(language=language)

    def topic_clustering(self, stemmed_phrases: list):
        """ grouping given phrases to topic based on HAC by there tokens

         Parameter
        --------------------
        stemmed_phrases: list
            list of stemmed keywords

         Return
        --------------------
        grouped_phrase: list
            grouped keywords
        """
        unique_token = set(chain(*[p.split() for p in stemmed_phrases]))
        token_to_id = dict([(t, i) for i, t in enumerate(unique_token)])

        # convert phrases to vector spanned by unique token
        matrix = np.zeros((len(stemmed_phrases), len(unique_token)))
        for n, p in enumerate(stemmed_phrases):
            indices = [token_to_id[_p] for _p in set(p.split())]
            matrix[n, indices] = 1

        # calculate distance
        distance = pdist(matrix, 'jaccard')
        # compute the clusters
        links = linkage(distance, method=self.__linkage_method)
        # get cluster id: len(clusters) == len(stemmed_phrases)
        clusters = fcluster(links, t=self.__clustering_threshold, criterion='distance')
        grouped_phrase = [np.array(stemmed_phrases)[clusters == c_id].tolist() for c_id in set(clusters)]
        return grouped_phrase

    def build_graph(self, document: str):
        """ Build basic graph with Topic """

        # convert phrase instance
        phrase_instance, stemmed_tokens = self.phrase_constructor.tokenize_and_stem_and_phrase(document)
        if len(phrase_instance) < 2:
            # at least 2 phrase are needed to extract keywords
            return None

        # group phrases as topic classes
        grouped_phrases = self.topic_clustering(list(phrase_instance.keys()))
        group_id = list(range(len(grouped_phrases)))

        # initialize graph instance
        graph = nx.Graph()

        # add nodes
        graph.add_nodes_from(group_id)

        def offset_distance(a: list, b: list):
            return 1/(a[0] - b[1] + 1) if a[0] > b[0] else 1/(b[0] - a[1] + 1)

        # add edges with weight
        for id_x, id_y in combinations(group_id, 2):
            # list of offset
            x_offsets = list(chain(*[phrase_instance[_x_phrase]['offset'] for _x_phrase in grouped_phrases[id_x]]))
            y_offsets = list(chain(*[phrase_instance[_y_phrase]['offset'] for _y_phrase in grouped_phrases[id_y]]))

            weight = sum([offset_distance(_x, _y) for _x, _y in product(x_offsets, y_offsets)])
            graph.add_edge(id_x, id_y, weight=weight)
        return graph, phrase_instance, grouped_phrases, len(stemmed_tokens)

    def run_pagerank(self, graph):
        node_score = nx.pagerank(G=graph, alpha=self.__random_prob, tol=self.__tol, weight='weight')
        return node_score

    def get_keywords(self, document: str, n_keywords: int = 10):
        """ Get keywords

         Parameter
        ------------------
        document: str
        n_keywords: int

         Return
        ------------------
        a list of dictionary consisting of 'stemmed', 'pos', 'raw', 'offset', 'count'.
        eg) {'stemmed': 'grid comput', 'pos': 'ADJ NOUN', 'raw': ['grid computing'], 'offset': [[11, 12]], 'count': 1}
        """
        # make graph and get data structure for candidate phrase
        output = self.build_graph(document)
        if output is None:
            return []

        graph, phrase_instance, grouped_phrases, original_sentence_token_size = output
        # pagerank to get score for individual word (sum of score will be one)
        node_score = self.run_pagerank(graph)

        # combine score to get score of phrase
        def aggregation(group_id, phrases):
            first_phrase_id = np.argmin([phrase_instance[p]['offset'][0][0] for p in phrases])
            phrase = phrases[first_phrase_id]
            return phrase, node_score[group_id]

        phrase_score = [aggregation(group_id, phrases) for group_id, phrases in enumerate(grouped_phrases)]

        # sorting
        phrase_score_sorted_list = sorted(phrase_score, key=lambda key_value: key_value[1], reverse=True)
        count_valid = min(n_keywords, len(phrase_score_sorted_list))

        def modify_output(stem, score):
            tmp = phrase_instance[stem]
            tmp['score'] = score
            tmp['n_source_tokens'] = original_sentence_token_size
            return tmp

        val = [modify_output(stem, score) for stem, score in phrase_score_sorted_list[:count_valid]]
        return val
